import os
import sys
import shutil
import platform
import git as gitpython
import venv as venvpython
import subprocess
from distutils.errors import DistutilsFileError
from distutils.dir_util import copy_tree
import typer
from tabulate import tabulate
from typing import Optional

# local imports
from trex import utils, meta

# app configuration
app = typer.Typer()
APP_NAME = meta.APP_NAME
APP_VERSION = meta.APP_VERSION


# ======================================================================================
# TREX TODOS

# yaaay, it's empty!

# ======================================================================================

@app.command()
def version():
    dir_path, config_path, templates_path = utils.get_app_dir()
    logo_text = typer.style("""
                 _
                | |_ _ __ _____  __
                | __| '__/ _ \ \/ /
                | |_| | |  __/>  <
                 \__|_|  \___/_/\_\ 
    """, fg=typer.colors.GREEN, bold=True)

    logo_line = typer.style(f"""
    ----------------------------------------------
        Version {APP_VERSION}    🦖    by berrysauce
    ----------------------------------------------
    """, fg=typer.colors.BRIGHT_GREEN, bold=True)

    more_info = typer.style("""
        Honey, it's the Templatosaurus Rex!
        Docs and more at berrysauce.me/trex
    """, fg=typer.colors.WHITE, bold=False)

    path_info = typer.style(f"""
    trex is located at:
    {dir_path}
    """, fg=typer.colors.BRIGHT_BLACK, bold=False)

    typer.echo(logo_text+logo_line+more_info+path_info)


@app.command()
def create(name: str):
    utils.print_start()

    utils.print_working("Getting template location")
    location = str(os.getcwd())

    utils.print_working("Adding template to storage")
    res = utils.add_template(name, {"location": location})

    if res:
        utils.print_done(f"Added '{name}' as template")
    else:
        utils.print_warn("A template with this name already exists")
    utils.show_tip(f"Use 'trex make {name}' to create a new directory from the template")


@app.command()
def remove(name: str):
    utils.print_start()
    utils.print_working("Removing template")

    if utils.remove_template(name) is True:
        utils.print_done(f"{name} template was removed")
    else:
        utils.print_warn(f"{name} template doesn't exist")


@app.command()
def make(name: str, target: str, git: bool = typer.Option(False), venv: bool = typer.Option(False), installdeps: bool = typer.Option(False)):
    utils.print_start()
    utils.print_working(f"Making directory from {name} template")

    utils.print_working("Fetching template data")
    res = utils.get_template(name)
    if res is None:
        utils.print_error("Template not found")
        return

    utils.print_working("Moving files around")
    try:
        destination = str(os.getcwd()) + "/" + target
        utils.print_working("Creating target directory")
        os.mkdir(destination)
    except FileExistsError:
        utils.print_error("File or directory already exists")
        return

    try:
        copy_tree(res["location"], destination)
    except DistutilsFileError:
        utils.print_error("Template directory doesn't seem to exist anymore")
        utils.print_working("Removing template")
        utils.print_working("Removing target directory")

        os.rmdir(destination)
        if utils.remove_template(name) is True:
            utils.print_done(f"{name} was removed")
        return

    if venv is True:
        # for documentation: https://docs.python.org/3/library/venv.html#venv.create
        # and: https://docs.python.org/3/library/venv.html#venv.EnvBuilder
        utils.print_working("Initializing virtual environment (venv) as .venv")
        # upgrade_deps=False since Python 3.9
        try:
            venvpython.create(destination+"/.venv", system_site_packages=False, clear=False, symlinks=False, with_pip=True, prompt=None, upgrade_deps=False)
        except Exception as exception:
            utils.print_error("Exception while initializing venv as .venv: \n" + str(exception))
            utils.print_working("Removing target directory")
            shutil.rmtree(destination)
            utils.print_done("Removed target directory")
            return

    if installdeps is True:
        utils.print_working("Installing dependencies from requirements.txt")
        utils.print_warn("Sorry, but requirements installation is not yet supported. You'll have to do that yourself after activating your newly created venv")
        """
        MIGHT WORK, CHECK LATER
        
        try:
            os.chdir(destination)
            current_system = platform.system()
            utils.print_working("Activating venv")
            if current_system == "Windows":
                os.system(f"source {destination}/venv/bin/activate")
            elif current_system == "Linux":
                os.system(f"source {destination}/venv/bin/activate")
            elif current_system == "Darwin":
                # or MacOS
                os.system(f"source {destination}/venv/bin/activate")
                
            utils.print_working("Installing requirements")
            typer.echo(utils.terminal_width * "-")
            subprocess.check_call([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])
            typer.echo(utils.terminal_width * "-")
        except subprocess.CalledProcessError:
            typer.echo(utils.terminal_width * "-")
            utils.print_warn("requirements.txt not found (no requirements installed)")
        """

    if git is True:
        utils.print_working("Initializing git repo")
        gitpython.Repo.init(destination)

    utils.print_done(f"Created {target} from {name}")


@app.command()
def all():
    res = utils.get_template(name=None)
    if res is None:
        utils.print_warn("Create a template first")
        return

    typer.secho("All available templates:", fg=typer.colors.BRIGHT_GREEN)
    head = ["Name", "Location"]
    data = []
    key_list = list(res.keys())
    values_list = list(res.values())

    for i in range(len(key_list)):
        data.append([key_list[i], values_list[i]["location"]])

    typer.echo(tabulate(data, headers=head, tablefmt="grid"))
    utils.show_tip(f"Use 'trex make <template name>' to create a new directory from the template")


@app.command()
def reset(force: bool = typer.Option(False)):
    if force is False:
        warning = typer.style(" ⚠️ WARNING ", fg=typer.colors.BRIGHT_YELLOW, bold=True)
        typer.confirm(warning + " Do you really want to reset trex and delete all its data?", abort=True)

    dir_path, config_path, templates_path = utils.get_app_dir()
    try:
        utils.print_working("Deleting config storage")
        os.remove(config_path)
    except FileNotFoundError or NotADirectoryError:
        pass

    try:
        utils.print_working("Deleting templates storage")
        os.remove(templates_path)
    except FileNotFoundError or NotADirectoryError:
        pass

    try:
        utils.print_working("Deleting trex storage directory")
        os.rmdir(dir_path)
    except FileNotFoundError or NotADirectoryError:
        pass
    utils.print_done("Reset complete")


@app.command()
def config(key: str, enable: bool = typer.Option(False), disable: bool = typer.Option(False)):
    utils.print_start()

    if key not in meta.config_options:
        utils.print_warn(f"{key} is not a valid config option")
        return

    if enable:
        value = True
    elif disable:
        value = False
    else:
        utils.print_warn("You need to provide --enable or --disable")
        return

    utils.print_working("Updating config")
    utils.add_config(key, data=value)
    utils.print_done(f"{key} set to {value}")


import os
from pathlib import Path
import typer
import json
import random

# local imports
from trex import meta

APP_NAME = meta.APP_NAME
APP_VERSION = meta.APP_VERSION

terminal_width, terminal_height = os.get_terminal_size()


"""
=============================================================
                   DIRECTORY HANDLING
=============================================================
"""

def get_app_dir():
    app_dir = typer.get_app_dir(APP_NAME)
    config_path: Path = Path(app_dir) / "config.json"
    templates_path: Path = Path(app_dir) / "templates.json"
    return Path(app_dir), config_path, templates_path

def check_app_dir():
    dir_path, config_path, templates_path = get_app_dir()
    try:
        os.mkdir(dir_path)
    except FileExistsError:
        pass
    return


"""
=============================================================
                TEMPLATE CONFIG HANDLING
=============================================================
"""

def add_template(name: str, data: dict):
    dir_path, config_path, templates_path = get_app_dir()
    check_app_dir()

    try:
        with open(templates_path, "r") as f:
            stored_config = f.read()
        stored_config = json.loads(stored_config)
    except FileNotFoundError:
        stored_config = {}

    try:
        test_exists = stored_config[name]
        return False
    except KeyError:
        pass

    stored_config[name] = data
    with open(templates_path, "w") as f:
        f.write(json.dumps(stored_config))

    return True


def get_template(name):
    dir_path, config_path, templates_path = get_app_dir()
    check_app_dir()

    try:
        with open(templates_path, "r") as f:
            stored_config = json.loads(f.read())

        if name is None:
            return stored_config
        else:
            return stored_config[name]
    except FileNotFoundError:
        return None
    except KeyError:
        return None


def remove_template(name: str):
    dir_path, config_path, templates_path = get_app_dir()
    check_app_dir()

    try:
        with open(templates_path, "r") as f:
            stored_config = json.loads(f.read())
        stored_config.pop(name)
    except FileNotFoundError:
        return None
    except KeyError:
        return None

    with open(templates_path, "w") as f:
        f.write(json.dumps(stored_config))

    return True




"""
=============================================================
                      CONFIG HANDLING
=============================================================
"""

def add_config(key: str, data):
    dir_path, config_path, templates_path = get_app_dir()
    check_app_dir()

    if key not in meta.config_options:
        return None

    try:
        with open(config_path, "r") as f:
            stored_config = f.read()
        stored_config = json.loads(stored_config)
    except FileNotFoundError:
        stored_config = {}

    stored_config[key] = data
    with open(config_path, "w") as f:
        f.write(json.dumps(stored_config))

    return True


def get_config(key: str):
    dir_path, config_path, templates_path = get_app_dir()
    check_app_dir()

    try:
        with open(config_path, "r") as f:
            stored_config = json.loads(f.read())
        return stored_config[key]
    except FileNotFoundError:
        return None
    except KeyError:
        return None


"""
=============================================================
                      TIP HANDLING
=============================================================
"""

def show_tip(msg: str):
    # only show tips when show_tips is True or doesn't exist
    show_tips = get_config("tips")
    if show_tips is not None and show_tips is False:
        return

    tip_title = typer.style(" TIP ", fg=typer.colors.BLACK, bg=typer.colors.WHITE, bold=True)
    tip = "\n" + tip_title + f" {msg}\n" + "Disable tips with 'trex config tips --disable'\n"
    typer.echo(tip)


"""
=============================================================
                      PRINT HANDLING
=============================================================
"""

def print_start():
    show_intro = get_config("intro")
    if show_intro is not None and show_intro is False:
        return
    working = random.choice(meta.working_strings)
    typer.secho("🦖 " + working, fg=typer.colors.BRIGHT_GREEN)

def print_warn(msg: str):
    warn_title = typer.style(" WARN! ", fg=typer.colors.BRIGHT_YELLOW, bold=True)
    typer.echo(warn_title + " " + msg)

def print_error(msg: str):
    warn_title = typer.style(" ERROR ", fg=typer.colors.BRIGHT_RED, bold=True)
    typer.echo(warn_title + " " + msg)

def print_working(msg: str):
    warn_title = typer.style(" ..... ", fg=typer.colors.BRIGHT_CYAN, bold=True)
    typer.echo(warn_title + " " + msg)

def print_done(msg: str):
    warn_title = typer.style(" DONE! ", fg=typer.colors.BRIGHT_GREEN, bold=True)
    typer.echo(warn_title + " " + msg)



"""
trex meta.py
Includes variables with meta info to import in other packages

CHANGE VALUES HERE WHEN UPDATED (VERSION)
"""

# App info
APP_NAME = "trex"
APP_VERSION = "0.1.4"

# Server info
TREX_SERVER = "trexapi.deta.dev"

# Config options
config_options = [
    "intro",
    "tips"
]

# "Getting to work" strings
working_strings = [
    "I'm on the job",
    "Let's roar!",
    "Roooaaar!",
    "Never seen a Templatosaurus Rex before?",
    "Getting to work",
    "I'm working, chill!",
    "So much to do today...",
    "How many times did you use this template again?",
    "I'm a Templatosaurus Rex",
    "As green as your terminal font, or not...",
    "I'm like a Genie, you only need one command to call me!"
]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trex']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0', 'tabulate>=0.8.9,<0.9.0', 'typer>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['trex = trex.main:app']}

setup_kwargs = {
    'name': 'trex',
    'version': '0.1.4',
    'description': 'Templatosaurus Rex 🦖 Create and manage project templates in your terminal',
    'long_description': '<img alt="trex Logo" src="https://cdn.berrysauce.me/assets/trex-banner.jpg">\n<p align="center"><strong>Templatosaurus Rex: Create and manage project templates in your terminal 🦖</strong></p>\n<p align="center">\n    <img alt="GitHub repo size" src="https://img.shields.io/github/repo-size/berrysauce/trex?label=size">\n    <img alt="GitHub last commit" src="https://img.shields.io/github/last-commit/berrysauce/trex">\n    <img alt="GitHub CodeQL" src="https://github.com/berrysauce/trex/actions/workflows/codeql-analysis.yml/badge.svg">\n    <img alt="PyPI - Downloads" src="https://img.shields.io/pypi/dm/trex?label=PyPi%20downloads">\n    <img alt="PyPI" src="https://img.shields.io/pypi/v/trex">\n    <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/trex">\n    <img alt="PyPI - Status" src="https://img.shields.io/pypi/status/trex">\n</p>\n\n---\n\n## 🦖 What is trex?\ntrex is a template manager in the form of a CLI app. You can create, organize, and clone-from template directories. It doesn’t matter if the template is a directory on your machine or a GitHub repository: with trex you only need a few seconds to actually get to coding. Not only that, but trex is really easy to use.\n\n## 🚧 Features and Roadmap\n- [x] Organise templates (create, remove, list)\n- [x] Make project from local directory template\n- [ ] Make project from GitHub template  \n- [x] Create git repository on creation\n- [ ] Run scripts automatically on creation\n- [ ] Create virtualenv automatically and import requirements on creation\n\n→ Do you have any feature requests? [Submit them here](https://github.com/berrysauce/trex/issues).\n\n## Install trex\n**Make sure you have Python 3.9 or above installed on your system.**\n\ntrex is available on PyPi and can be installed via pip with the following command:\n```\npip install trex\n```\nCheck if the installation was successful:\n```\ntrex version\n```\n\n## 📜 License\n\ntrex: A modern and intuitive templating CLI\nCopyright (C) 2022 berrysauce (Paul Haedrich)\n\nThis program is free software: you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program.  If not, see <https://www.gnu.org/licenses/>.\n\nFor legal questions, contact legal[at]berrysauce[dot]me.\n\n<a href="https://deta.sh/?ref=basketball" target="_blank"><img src="https://cdn.berrysauce.me/assets/deta-sponsor-banner.jpg" alt="Sponsored by Deta"></a>\n',
    'author': 'Paul Haedrich',
    'author_email': 'paul@berrysauce.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://berrysauce.me/trex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
